/* crypt-gpgme.c - GPGME based crypto operations
 * Copyright (C) 1996,1997 Michael R. Elkins <me@cs.hmc.edu>
 * Copyright (C) 1998,1999,2000 Thomas Roessler <roessler@guug.de>
 * Copyright (C) 2001  Thomas Roessler <roessler@guug.de>
 *                     Oliver Ehli <elmy@acm.org>
 * Copyright (C) 2002 g10 Code GmbH
 *
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 * 
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 * 
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.
 */
/*
  Module interface description:


  mutt_is_application_pgp 
  mutt_is_application_smime

    Classify a given body.  Used in set_copy_flags, mutt_can_decode,
    mutt_body_handler, mutt_needs_mailcap, mutt_is_text_part,
    mutt_view_attachments, mutt_make_message_attach,
    mutt_write_mime_body,

  mutt_is_multipart_signed
  
    Classify a given body. Used in mutt_prepare_template

  mutt_is_multipart_encryped

    Classify a given body. Used in set_copy_flags,
    mutt_prepare_template, mutt_gen_attach_list,
    mutt_view_attachments.


  mutt_signed_handler

    MIME handler for all multipart/signed.  Used by mutt_body_handler
    for this type.  This catches S/MIEm as well as PGP/MIME.


  pgp_application_pgp_handler

    MIME handler for classic application/pgp. Used by
    mutt_body_handler for text/plain and application/.*  if
    mutt_is_application_pgp returns true.  This handles both, signed
    and encrypted messages.


  pgp_encrypted_handler

    MIME handler for PGP/MIME encrypted. Used by mutt_body_handler for
    multipart/encrypted when the protocol is
    applciation/gpg-encrypted.


  smime_encrypted_handler

    MIME handler for application/smime.  Used by mutt_body_handler for
    application/.* if mutt_is_applicatoion_smime returns true.  This
    handles encrypted and opaque signed S/MIME messages.


  pgp_decrypt_mime

    Decrypts a PGP/MIME message and returns a new stream and body.
    Used by _mutt_copy_message, mutt_prepare_template,
    mutt_view_attachments.  Locally used by crypt_decrypt_mime.

  smime_decrypt_mime

    Decrypts a S/MIME message and returns a new stream and body.  Used
    by _mutt_copy_message, mutt_view_attachments.  Locally used by
    crypt_decrypt_mime.  
    FIXME: This is not used by mutt_prepare_template - check why.

  crypt_decrypt_mime

    Wrapper around pgp_descrypt_mime and smime_descrypt_mime.  Used by
    mutt_view_attachments.  We should check whether the other uses of
    in this function are at all used when build with GPGME.

  pgp_check_traditional

    Used in _mutt_check_traditional_pgp, mutt_view_attachments.


  */



#include "mutt.h"

#ifdef HAVE_GPGME

#include "mutt_curses.h"
#include "crypt.h"
#include "mime.h"
#include "copy.h"

#include <sys/wait.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/stat.h>
#include <errno.h>
#include <ctype.h>

#ifdef HAVE_LOCALE_H
#include <locale.h>
#endif
#ifdef HAVE_LANGINFO_D_T_FMT
#include <langinfo.h>
#endif

#ifdef HAVE_SYS_TIME_H
# include <sys/time.h>
#endif

#ifdef HAVE_SYS_RESOURCE_H
# include <sys/resource.h>
#endif

#include "crypt.h"

/* safe_free is evil in that a cast is required and one often forgets
   to pass the address and not just the pointer.  This scheme bypasses
   all type checking and thus is not reqyure.  And free() is suppored
   to be a noop when a NULL pointer is passed.  A macro is the correct
   solution. */
#undef FREE
#define FREE(a) do { free (a); a = NULL; } while (0)
#define safe_free(a) you_should_not_use_safe_free (a);

static int set_signer (GpgmeCtx ctx, int for_smime);
static char *find_keys (ADDRESS *to, ADDRESS *cc, ADDRESS *bcc,
                        unsigned int *r_application);
static int verify_one (BODY *sigbdy, STATE *s,
                       const char *tempfile, int is_smime);
static BODY *sign_message (BODY *a, int use_smime);
static BODY *encrypt_message_pgp (BODY *a, char *keylist, int sign);
static BODY *encrypt_message_smime (BODY *a, char *certlist);

/* Create a new gpgme context and return it.  With FOR_SMIME set to
   true, the protocol of the context is set to CMS. */
GpgmeCtx create_gpgme_context (int for_smime)
{
  GpgmeError err;
  GpgmeCtx ctx;

  err = gpgme_new (&ctx);
  if (err)
    {
      mutt_error ("error creating gpgme context: %s\n", gpgme_strerror (err));
      sleep (2);
      mutt_exit (1);
    }

  if (for_smime) 
    {
      err = gpgme_set_protocol (ctx, GPGME_PROTOCOL_CMS);
      if (err)
        {
          mutt_error ("error enabling CMS protocol: %s\n",
                      gpgme_strerror (err));
          sleep (2);
          mutt_exit (1);
        }
    }

  return ctx;
}

/* Create a new gpgme data object.  This is a wrapper to die on
   error. */
GpgmeData create_gpgme_data (void)
{
  GpgmeError err;
  GpgmeData data;

  err = gpgme_data_new (&data);
  if (err) 
    {
      mutt_error ("error creating gpgme data object: %s\n",
                  gpgme_strerror (err));
      sleep (2);
      mutt_exit (1);
    }
  return data;
}



/* Given a message body M with already parsed headers, figure out
   whether the message is PGP or S/MIME encoded (signed or encrypted).
   Return a flag indicating the type of encoding or 0 if it is not
   such a message. Note, that this function does not detect a PGP/MIME
   message. */
int mutt_is_crypto (BODY *m)
{
  int t = 0;
  char *p;

  if (!m)
    ;
  else if (m->is_signed_data)
    t |= SMIMESIGN|SMIMEOPAQUE; 
  else if (m->type == TYPEAPPLICATION && m->subtype )
    { 
      int try_name = 0;

      if (!ascii_strcasecmp (m->subtype, "pgp")
          || !ascii_strcasecmp (m->subtype, "x-pgp-message"))
        {
          if ((p = mutt_get_parameter ("x-action", m->parameter))
              && (!ascii_strcasecmp (p, "sign")
                  || !ascii_strcasecmp (p, "signclear")))
            t |= PGPSIGN;
          
          if ((p = mutt_get_parameter ("format", m->parameter))
              && !ascii_strcasecmp (p, "keys-only"))
            t |= PGPKEY;
          
          if(!t)
            t |= PGPENCRYPT;  /* best guess we can do. */
        }
      else if (!ascii_strcasecmp (m->subtype, "pgp-signed"))
        t |= PGPSIGN;
      else if (!ascii_strcasecmp (m->subtype, "pgp-keys"))
        t |= PGPKEY;
      else if (!ascii_strcasecmp (m->subtype, "pkcs7-mime")
               ||! ascii_strcasecmp (m->subtype, "x-pkcs7-mime"))
        {
          if ((p = mutt_get_parameter ("smime-type", m->parameter)))
            {
              if (!ascii_strcasecmp (p, "enveloped-data"))
                t |= SMIMEENCRYPT;
              else if (!ascii_strcasecmp (p, "signed-data"))
                t |= SMIMESIGN|SMIMEOPAQUE;
            }
          else if (!ascii_strcasecmp (m->description,
                                      "S/MIME Encrypted Message"))
            t |= SMIMEENCRYPT; /* for Netscape 4.7 */
          else
            try_name = 1;
        }
      else if (!ascii_strcasecmp (m->subtype, "octet-stream"))
        try_name = 1;

      if (try_name)
        { /* Try to figure the type from other info. */
          p = mutt_get_parameter ("name", m->parameter);
          if (!p)
            p = m->d_filename;
          if (!p)
            p = m->filename;
          if (p)
            {
              int len = mutt_strlen (p) - 4;
              if (len > 0 && p[len] == '.')
                {
                  len++;
                  if (!ascii_strcasecmp (p+len, "p7m"))
                    t |= SMIMEENCRYPT;
                  else if (!ascii_strcasecmp (p+len, "p7s"))
                    t |= SMIMESIGN|SMIMEOPAQUE;
                }
            }
        }
    }
  else if (m->type == TYPETEXT && m->subtype
           && !ascii_strcasecmp ("plain", m->subtype))
    {
      if (((p = mutt_get_parameter ("x-mutt-action", m->parameter))
           || (p = mutt_get_parameter ("x-action", m->parameter)) 
           || (p = mutt_get_parameter ("action", m->parameter)))
          && !ascii_strncasecmp ("pgp-sign", p, 8))
        t |= PGPSIGN;
      else if (p && !ascii_strncasecmp ("pgp-encrypt", p, 11))
        t |= PGPENCRYPT;
    }
  else if  (m->type == TYPEMULTIPART && m->subtype
            && !ascii_strcasecmp (m->subtype, "encrypted"))
    {
      p = mutt_get_parameter ("protocol", m->parameter);
      if (p && !ascii_strcasecmp (p, "application/pgp-encrypted"))
        t = PGPENCRYPT;
    }
  else if  (m->type == TYPEMULTIPART && m->subtype
            && !ascii_strcasecmp (m->subtype, "signed"))
    {
      p = mutt_get_parameter ("protocol", m->parameter);
      if (!p)
        ;
      else if (!mutt_strcasecmp (p, "multipart/mixed"))
        t = SIGN;
      else if (!mutt_strcasecmp (p, "application/pgp-signature"))
        t = PGPSIGN;
      else if (!mutt_strcasecmp (p, "application/pkcs7-signature"))
        t = SMIMESIGN;
      else if (!mutt_strcasecmp (p, "application/x-pkcs7-signature"))
        t = SMIMESIGN;
    }


  return t;
}

int mutt_is_application_pgp (BODY *m)
{
  int t = 0;
  if (m && m->type != TYPEMULTIPART) 
    {
      switch ((t= mutt_is_crypto (m)))
        {
        case PGPENCRYPT:
        case PGPSIGN:
          break;
        default:
          t = 0;
          break;
        }
    }
  return t;
}

int mutt_is_application_smime (BODY *m)
{
  int t = 0;
  if (m && m->type != TYPEMULTIPART) 
    {
      switch ((t= mutt_is_crypto (m)))
        {
        case SMIMESIGN:
        case (SMIMESIGN|SMIMEOPAQUE):
        case SMIMEENCRYPT:
          break;
        default:
          t = 0;
          break;
        }
    }
  return t;
}

int mutt_is_multipart_encrypted (BODY *m)
{
  int t = 0;
  if (m && m->type == TYPEMULTIPART) 
    {
      switch ((t= mutt_is_crypto (m)))
        {
        case PGPENCRYPT:
          break;
        default:
          t = 0;
          break;
        }
    }
  return t;
}

int mutt_is_multipart_signed (BODY *m)
{
  int t = 0;
  if (m && m->type == TYPEMULTIPART) 
    {
       switch ((t=mutt_is_crypto (m)))
         {
         case SIGN:
         case PGPSIGN:
         case SMIMESIGN:
         case (SMIMEOPAQUE|SMIMESIGN):
           break;
         default:
           t = 0;
           break;
         }
    }
  return t;
}


/* Check out the type of encryption used and set the cached status
   values if there are any. */
int crypt_query (BODY *m)
{
  int t = 0;

  if (m->type == TYPEAPPLICATION)
    {
      t |= mutt_is_crypto (m);
      if (t && m->goodsig) 
        t |= GOODSIGN;
    }
  else if (m->type == TYPETEXT)
    {
      t |= mutt_is_application_pgp (m);
      if (t && m->goodsig)
        t |= GOODSIGN;
    }
  else if (m->type == TYPEMULTIPART)
    {
      t |= mutt_is_multipart_encrypted(m);
      t |= mutt_is_multipart_signed (m);

      if (t && m->goodsig)
        t |= GOODSIGN;
    }
  
  if (m->type == TYPEMULTIPART || m->type == TYPEMESSAGE)
  {
    BODY *p;
    
    for (p = m->parts; p; p = p->next)
      t |= crypt_query (p) & ~GOODSIGN;
  }
  
  return t;
}



static void print_time(time_t t, STATE *s)
{
  char p[STRING];

  setlocale (LC_TIME, "");
#ifdef HAVE_LANGINFO_D_T_FMT
  strftime (p, sizeof (p), nl_langinfo (D_T_FMT), localtime (&t));
#else
  strftime (p, sizeof (p), "%c", localtime (&t));
#endif
  setlocale (LC_TIME, "C");
  state_attach_puts (p, s);
}

/* print the current time to avoid spoofing of the signature output.
   This is not used with gpgme */
void crypt_current_time(STATE *s, char *app_name)
{
  
}



void crypt_forget_passphrase (void)
{
  /* fixme: notify gpg-agent to forget the passphrases.  Note that
     this function is not anymore so important as we qdon't have to care
     about passphrases in Mutt at all */
  mutt_message _("Passphrase(s) are now handled by gpg-agent");
}


void smime_getkeys (ENVELOPE *env)
{
  /* Dummy - we don't need this for gpgme. */
}


int crypt_valid_passphrase(int flags)
{
  return 1; /* always valid because it is handled by gpg-agent */
}


int mutt_protect (HEADER *msg, char *keylist)
{
  BODY *pbody = NULL, *tmp_pbody = NULL;
  BODY *tmp_smime_pbody = NULL;
  BODY *tmp_pgp_pbody = NULL;
  int flags = msg->security;

#if 0/* traditional encryption */
  int traditional = 0;
  /* FIXME: take code from crypt.c */
#endif
  
  if (!isendwin ())
    mutt_endwin (NULL);

  tmp_smime_pbody = msg->content;

  if (msg->security & SIGN)
    {
      if (msg->security & APPLICATION_SMIME)
        {
          if (!(tmp_pbody = sign_message (msg->content, 1)))
            return -1;
          pbody = tmp_smime_pbody = tmp_pbody;
        }

      if ((msg->security & APPLICATION_PGP) &&
          (!(flags & ENCRYPT) || option (OPTPGPRETAINABLESIG)))
        {
          if (!(tmp_pbody = sign_message (msg->content, 0)))
            return -1;
          
          flags &= ~SIGN;
          pbody = tmp_pgp_pbody = tmp_pbody;
        }

      if ((msg->security & APPLICATION_SMIME) &&
          (msg->security & APPLICATION_PGP))
        {
          /* here comes the draft ;-) */
        }
    }


  if (msg->security & ENCRYPT)
    {
      if (msg->security & APPLICATION_SMIME)
        {
          if (!(tmp_pbody = encrypt_message_smime (tmp_smime_pbody, keylist)))
            {
              /* signed ? free it! */
              return (-1);
            }
          /* free tmp_body if messages was signed AND encrypted ... */
          if (tmp_smime_pbody != msg->content && tmp_smime_pbody != tmp_pbody)
            {
              /* detatch and dont't delete msg->content,
                 which tmp_smime_pbody->parts after signing. */
              tmp_smime_pbody->parts = tmp_smime_pbody->parts->next;
              msg->content->next = NULL;
              mutt_free_body (&tmp_smime_pbody);
            }
          pbody = tmp_pbody;
        }

      if (msg->security & APPLICATION_PGP)
        {
          if (!(pbody = encrypt_message_pgp (tmp_pgp_pbody?
                                             tmp_pgp_pbody : msg->content,
                                             keylist,
                                             flags & SIGN)))
            {
              /* did we perform a retainable signature? */
              if (flags != msg->security)
                {
                  /* remove the outer multipart layer */
                  msg->content = mutt_remove_multipart (msg->content);
                  /* get rid of the signature */
                  mutt_free_body (&msg->content->next);
                }
              
              return (-1);
            }

          /* destroy temporary signature envelope when doing
             retainable signatures.  */
          if (flags != msg->security)
            {
              mutt_remove_multipart (msg->content);
              mutt_free_body (&msg->content->next);
            }
        }
    }

  if(pbody)
    msg->content = pbody;
  
  return 0;
}



/* Write the message body/part A described by state S to a the given
   TEMPFILE.  fixme: There is no need to use the tempfiles with gpgme
   but for now we go with this. */
int crypt_write_signed(BODY *a, STATE *s, const char *tempfile)
{
  FILE *fp;
  int c;
  short hadcr;
  size_t bytes;

  if (!(fp = safe_fopen (tempfile, "w")))
  {
    mutt_perror (tempfile);
    return -1;
  }
      
  fseek (s->fpin, a->hdr_offset, 0);
  bytes = a->length + a->offset - a->hdr_offset;
  hadcr = 0;
  while (bytes > 0)
  {
    if ((c = fgetc (s->fpin)) == EOF)
      break;
    
    bytes--;
    
    if  (c == '\r')
      hadcr = 1;
    else 
    {
      if (c == '\n' && !hadcr)
	fputc ('\r', fp);
      
      hadcr = 0;
    }
    
    fputc (c, fp);
    
  }
  fclose (fp);

  return 0;
}


void convert_to_7bit (BODY *a)
{
  while (a)
  {
    if (a->type == TYPEMULTIPART)
    {
      if (a->encoding != ENC7BIT)
      {
        a->encoding = ENC7BIT;
	convert_to_7bit(a->parts);
      }
      else if (option (OPTPGPSTRICTENC))
	convert_to_7bit (a->parts);
    } 
    else if (a->type == TYPEMESSAGE &&
	     mutt_strcasecmp(a->subtype, "delivery-status"))
    {
      if(a->encoding != ENC7BIT)
	mutt_message_to_7bit (a, NULL);
    }
    else if (a->encoding == ENC8BIT)
      a->encoding = ENCQUOTEDPRINTABLE;
    else if (a->encoding == ENCBINARY)
      a->encoding = ENCBASE64;
    else if (a->content && a->encoding != ENCBASE64 &&
	     (a->content->from || (a->content->space && 
				   option (OPTPGPSTRICTENC))))
      a->encoding = ENCQUOTEDPRINTABLE;
    a = a->next;
  }
}



void crypt_extract_keys_from_messages (HEADER * h)
{
#if 0 
  /* FIXME take code from crypt.c */
#endif
}



/* Do a quick check to make sure that we can find all of the
   encryption keys if the user has requested this service. 
   Return the list of keys in KEYLIST. */
int crypt_get_keys (HEADER *msg, char **keylist)
{
  unsigned int app;

  set_option (OPTPGPCHECKTRUST);
  
  *keylist = NULL;
  
  if (msg->security & ENCRYPT)
    {
      if ( (msg->security & (APPLICATION_PGP|APPLICATION_SMIME)))
        {
          if (!(*keylist = find_keys (msg->env->to, msg->env->cc,
                                      msg->env->bcc, &app)))
            return -1;
          if (!app)
            {
              FREE (keylist);
              mutt_error (_("Error: can't mix key types"));
              return -1;
            }
          unset_option (OPTPGPCHECKTRUST);
          msg->security &= ~(APPLICATION_PGP|APPLICATION_SMIME);
          if (app & APPLICATION_PGP)
            msg->security |= APPLICATION_PGP;
          if (app & APPLICATION_SMIME)
            msg->security |= APPLICATION_SMIME;
        }
    }
    
  return 0;
}



static void crypt_fetch_signatures (BODY ***signatures, BODY *a, int *n)
{
  for (; a; a = a->next)
  {
    if (a->type == TYPEMULTIPART)
      crypt_fetch_signatures (signatures, a->parts, n);
    else
    {
      if((*n % 5) == 0)
	safe_realloc ((void **) signatures, (*n + 6) * sizeof (BODY **));

      (*signatures)[(*n)++] = a;
    }
  }
}


/* This routine attempts to find the keyids of the recipients of a
   message.  It returns NULL if any of the keys can not be found.  */
static char *find_keys (ADDRESS *to, ADDRESS *cc, ADDRESS *bcc,
                        unsigned int *r_application)
{
  char *keyID, *keylist = NULL, *t;
  size_t keylist_size = 0;
  size_t keylist_used = 0;
  ADDRESS *tmp = NULL, *addr = NULL;
  ADDRESS **last = &tmp;
  ADDRESS *p, *q;
  int i;
  crypt_key_t *k_info, *key;
  const char *fqdn = mutt_fqdn (1);

  *r_application = APPLICATION_PGP|APPLICATION_SMIME;
  
  for (i = 0; i < 3; i++) 
    {
      switch (i)
        {
        case 0: p = to; break;
        case 1: p = cc; break;
        case 2: p = bcc; break;
        default: abort ();
        }
      
      *last = rfc822_cpy_adr (p);
      while (*last)
        last = &((*last)->next);
    }
  
  if (fqdn)
    rfc822_qualify (tmp, fqdn);
  
  tmp = mutt_remove_duplicates (tmp);
  
  for (p = tmp; p ; p = p->next)
    {
      char buf[LONG_STRING];
      
      q = p;
      k_info = NULL;
      
      if ((keyID = mutt_crypt_hook (p)) != NULL)
        {
          int r;
          snprintf (buf, sizeof (buf), _("Use keyID = \"%s\" for %s?"),
                    keyID, p->mailbox);
          if ((r = mutt_yesorno (buf, M_YES)) == M_YES)
            {
              /* check for e-mail address */
              if ((t = strchr (keyID, '@')) && 
                  (addr = rfc822_parse_adrlist (NULL, keyID)))
                {
                  if (fqdn)
                    rfc822_qualify (addr, fqdn);
                  q = addr;
                }
              else
                k_info = crypt_getkeybystr (keyID, KEYFLAG_CANENCRYPT, 
                                            *r_application);
            }
          else if (r == -1)
            {
              FREE (keylist);
              rfc822_free_address (&tmp);
              rfc822_free_address (&addr);
              return NULL;
            }
        }

      if (k_info == NULL
          && (k_info = crypt_getkeybyaddr (q, KEYFLAG_CANENCRYPT)) == NULL)
        {
          snprintf (buf, sizeof (buf), _("Enter keyID for %s: "), q->mailbox);
          
          if ((key = crypt_ask_for_key (buf, q->mailbox,
                                        KEYFLAG_CANENCRYPT,
                                        *r_application)) == NULL)
            {
              FREE (keylist);
              rfc822_free_address (&tmp);
              rfc822_free_address (&addr);
              return NULL;
            }
        }
      else
        key = k_info;

      {
        const char *s = crypt_fpr (key);

        if (key->flags & KEYFLAG_ISX509)
          *r_application &= ~APPLICATION_PGP;
        if (!(key->flags & KEYFLAG_ISX509))
          *r_application &= ~APPLICATION_SMIME;
      
        keylist_size += mutt_strlen (s) + 4;
        safe_realloc ((void **)&keylist, keylist_size);
        sprintf (keylist + keylist_used, "%s0x%s",
                 keylist_used ? " " : "", /* __SPRINTF_CHECKED__ */ s);
      }
      keylist_used = mutt_strlen (keylist);
        
      crypt_free_key (&key);
      rfc822_free_address (&addr);
    }
  rfc822_free_address (&tmp);
  return (keylist);
}


/* Create a new GPGME Data object from the mail body A.  With CONVERT
   passed as true, the lines are converted to CR,LF if required.
   Return NULL on error or the GpgmeData object on success. */
static GpgmeData body_to_data_object (BODY *a, int convert)
{
  char tempfile[_POSIX_PATH_MAX];
  FILE *fptmp;
  int err = 0;
  GpgmeData data;
  
  mutt_mktemp (tempfile);
  fptmp = safe_fopen (tempfile, "w+");
  if (!fptmp)
    {
      mutt_perror (tempfile);
      return NULL;
    }

  mutt_write_mime_header (a, fptmp);
  fputc ('\n', fptmp);
  mutt_write_mime_body (a, fptmp);

  if (convert)
    {
      int c, hadcr = 0;
      unsigned char buf[1];

      data = create_gpgme_data ();
      rewind (fptmp);
      while ((c = fgetc (fptmp)) != EOF)
        {
          if  (c == '\r')
            hadcr = 1;
          else 
            {
              if (c == '\n' && !hadcr)
                {
                  buf[0] = '\r';
                  gpgme_data_write (data, buf, 1);
                }
                  
              hadcr = 0;
            }
          /* FIXME: This is quite suboptimal */
          buf[0] = c;
          gpgme_data_write (data, buf, 1);
        }
      fclose(fptmp);
    }
  else
    {
      fclose(fptmp);
      err = gpgme_data_new_from_file (&data, tempfile, 1);
    }
  unlink (tempfile);
  if (err) 
    {
      mutt_error ("error allocating data object: %s\n", gpgme_strerror (err));
      return NULL;
    }

  return data;
}

/* Create a GPGME data object from the stream FP but limit the object
   to LENGTH bytes starting at OFFSET bytes from the beginning of the
   file. */
static GpgmeData file_to_data_object (FILE *fp, long offset, long length)
{
  int err = 0;
  GpgmeData data;
  
  err = gpgme_data_new_from_filepart (&data, NULL, fp, offset, length);
  if (err) 
    {
      mutt_error ("error allocating data object: %s\n", gpgme_strerror (err));
      return NULL;
    }

  return data;
}

/* Write a GPGME data object to the stream FP. */
static int data_object_to_stream (GpgmeData data, FILE *fp)
{
  int err;
  char buf[4096], *p;
  size_t nread;

  err = gpgme_data_rewind (data);
  if (err)
    {
      mutt_error ("error rewinding data object: %s\n", gpgme_strerror (err));
      return -1;
    }

  while ( !(err = gpgme_data_read (data, buf, sizeof buf, &nread)) )
    {
      /* fixme: we are not really converting CRLF to LF but just
         skipping CR. Doing it correctly needs a more complex logic */
      for (p=buf; nread; p++, nread--)
        {
          if (*p != '\r')
            putc (*p, fp);
        }

     if (ferror (fp))
       {
         mutt_perror ("[tempfile]");
         return -1;
       }
    }
  if (err != GPGME_EOF)
    {
      mutt_error ("error reading data object: %s\n", gpgme_strerror (err));
      return -1;
    }
  return 0;
}

/* Copy a data object to a newly created temporay file and return that
   filename. Caller must free.  With RET_FP not NULL, don't close the
   stream but return it there. */
static char *data_object_to_tempfile (GpgmeData data, FILE **ret_fp)
{
  int err;
  char tempfile[_POSIX_PATH_MAX];
  FILE *fp;

  mutt_mktemp (tempfile);
  fp = safe_fopen (tempfile, "w+");
  if (!fp)
    {
      mutt_perror (tempfile);
      return NULL;
    }

  err = gpgme_data_rewind (data);
  if (!err)
    {
      char buf[4096];
      size_t nread;

      while ( !(err = gpgme_data_read (data, buf, sizeof buf, &nread)) )
        {
          if (fwrite (buf, nread, 1, fp) != 1)
            {
              mutt_perror (tempfile);
              fclose (fp);
              unlink (tempfile);
              return NULL;
            }
        }
    }
  if (ret_fp)
    rewind (fp);
  else
    fclose (fp);
  if (err != GPGME_EOF) 
    {
      mutt_error ("error reading data object: %s\n", gpgme_strerror (err));
      unlink (tempfile);
      fclose (fp);
      return NULL;
    }
  if (ret_fp)
    *ret_fp = fp;
  return safe_strdup (tempfile);
}


/* Create a GpgmeRecipientSet from the keys in the string KEYLIST.
   The keys must be space delimited. */
static GpgmeRecipients create_recipient_set (const char *keylist)
{
  int err;
  const char *s;
  char buf[100];
  int i;
  GpgmeRecipients rset;

  err = gpgme_recipients_new (&rset);
  if (err)
    {
      mutt_error ("error creating recipient set: %s\n", gpgme_strerror (err));
      return NULL;
    }
  
  s = keylist;
  do {
    while (*s == ' ')
      s++;
    for (i=0; *s && *s != ' ' && i < sizeof(buf)-1;)
      buf[i++] = *s++;
    buf[i] = 0;
    if (*buf)
      {
        err = gpgme_recipients_add_name (rset, buf);
        if (err)
          {
            mutt_error ("error setting recipient `%s': %s\n",
                        buf, gpgme_strerror (err));
            gpgme_recipients_release (rset);
            return NULL;
          }
      }
  } while (*s);

  return rset;
}


/* Encrypt the gpgme data object PLAINTEXT to the recipients in RSET
   and return an allocated filename to a temporary file containing the
   enciphered text.  With USE_SMIME set to true, the smime backend is
   used.  With COMBINED_SIGNED a PGP message is signed and
   encrypted.  Returns NULL in case of error */
static char *encrypt_gpgme_object (GpgmeData plaintext,
                                   GpgmeRecipients rset,
                                   int use_smime, int combined_signed)
{
  int err;
  GpgmeCtx ctx;
  GpgmeData ciphertext;
  char *outfile;

  ctx = create_gpgme_context (use_smime);
  if (!use_smime) 
      gpgme_set_armor (ctx, 1);

  ciphertext = create_gpgme_data ();

  if (combined_signed)
    {
      if (set_signer (ctx, use_smime))
        {
          gpgme_data_release (ciphertext);
          gpgme_release (ctx);
          return NULL;
        }
      err = gpgme_op_encrypt_sign (ctx, rset, plaintext, ciphertext);
    }
  else
    err = gpgme_op_encrypt (ctx, rset, plaintext, ciphertext);
  mutt_need_hard_redraw ();
  if (err)
    {
      mutt_error ("error encrypting data: %s\n", gpgme_strerror (err));
      gpgme_data_release (ciphertext);
      gpgme_release (ctx);
      return NULL;
    }

  gpgme_release (ctx);

  outfile = data_object_to_tempfile (ciphertext, NULL);
  gpgme_data_release (ciphertext);
  return outfile;
}

/* Find the "micalg" parameter from the last Gpgme operation on
   context CTX.  It is expected that this operation was a sign
   operation.  Return the algorithm name as a C string in buffer BUF
   which must have been allocated by the caller with size BUFLEN.
   Returns 0 on success or -1 in case of an error.  The return string
   is truncted to BUFLEN - 1. */
static int get_micalg (GpgmeCtx ctx, char *buf, size_t buflen)
{
  int i;
  const char *s = gpgme_get_op_info (ctx, 0);

  if (!buflen)
    return -1;
  *buf = 0;
  if (s && (s = strstr (s, "<micalg>")))
    {
      s += 8;
      for (i=0; *s && *s != '<' && *s != '>' && i < buflen-1;)
        buf[i++] = *s++;
      buf[i] = 0;
    }
  return *buf? 0:-1;
}

/* Make sure that the correct signer is set. Returns 0 on success. */
static int set_signer (GpgmeCtx ctx, int for_smime)
{
  char *signid = for_smime? SmimeSignAs: PgpSignAs;
  GpgmeError err;
  GpgmeCtx listctx;
  GpgmeKey key, key2;

  if (!signid || !*signid)
    return 0;

  listctx = create_gpgme_context (for_smime);
  err = gpgme_op_keylist_start (listctx, signid, 1);
  if (!err)
    err = gpgme_op_keylist_next (listctx, &key);
  if (err)
    {
      gpgme_release (listctx);
      mutt_error (_("secret key `%s' not found: %s\n"),
                  signid, gpgme_strerror (err));
      return -1;
    }
  err = gpgme_op_keylist_next (listctx, &key2);
  if (!err)
    {
      gpgme_key_release (key);
      gpgme_key_release (key2);
      gpgme_release (listctx);
      mutt_error (_("ambiguous specfication of secret key `%s'\n"),
                  signid);
      return -1;
    }
  gpgme_op_keylist_end (listctx);
  gpgme_release (listctx);

  gpgme_signers_clear (ctx);
  err = gpgme_signers_add (ctx, key);
  gpgme_key_release (key);
  if (err)
    {
      mutt_error (_("error setting secret key `%s': %s\n"),
                  signid, gpgme_strerror (err));
      return -1;
    }
  return 0;
}


/* Sign the MESSAGE in body A either using OpenPGP or S/MIME when
   USE_SMIME is passed as true.  Returns the new body or NULL on
   error. */
static BODY *sign_message (BODY *a, int use_smime)
{
  BODY *t;
  char *sigfile;
  int err = 0;
  char buf[100];
  GpgmeCtx ctx;
  GpgmeData message, signature;

  convert_to_7bit (a); /* Signed data _must_ be in 7-bit format. */

  message = body_to_data_object (a, 1);
  if (!message)
     return NULL;
  signature = create_gpgme_data ();

  ctx = create_gpgme_context (use_smime);
  if (!use_smime)
    gpgme_set_armor (ctx, 1);

  if (set_signer (ctx, use_smime))
    {
      gpgme_data_release (signature);
      gpgme_release (ctx);
      return NULL;
    }

  err = gpgme_op_sign (ctx, message, signature, GPGME_SIG_MODE_DETACH );
  mutt_need_hard_redraw ();
  gpgme_data_release (message);
  if (err)
    {
      gpgme_data_release (signature);
      gpgme_release (ctx);
      mutt_error ("error signing data: %s\n", gpgme_strerror (err));
      return NULL;
    }

  sigfile = data_object_to_tempfile (signature, NULL);
  gpgme_data_release (signature);
  if (!sigfile)
    {
      gpgme_release (ctx);
      return NULL;
    }

  t = mutt_new_body ();
  t->type = TYPEMULTIPART;
  t->subtype = safe_strdup ("signed");
  t->encoding = ENC7BIT;
  t->use_disp = 0;
  t->disposition = DISPINLINE;

  mutt_generate_boundary (&t->parameter);
  mutt_set_parameter ("protocol",
                      use_smime? "application/pkcs7-signature"
                               : "application/pgp-signature",
                      &t->parameter);
  /* Get the micalg from gpgme.  Old gpgme versions don't support this
     for S/MIME so we assume sha-1 in this case. */
  if (!get_micalg (ctx, buf, sizeof buf))
    mutt_set_parameter ("micalg", buf, &t->parameter);
  else if (use_smime)
    mutt_set_parameter ("micalg", "sha1", &t->parameter);
  gpgme_release (ctx);

  t->parts = a;
  a = t;

  t->parts->next = mutt_new_body ();
  t = t->parts->next;
  t->type = TYPEAPPLICATION;
  if (use_smime)
    {
      t->subtype = safe_strdup ("pkcs7-signature");
      mutt_set_parameter ("name", "smime.p7s", &t->parameter);
      t->encoding = ENCBASE64; 
      t->use_disp = 1;
      t->disposition = DISPATTACH;
      t->d_filename = safe_strdup ("smime.p7s");
    }
  else
    {
      t->subtype = safe_strdup ("pgp-signature");
      t->use_disp = 0;
      t->disposition = DISPINLINE;
      t->encoding = ENC7BIT;
    }
  t->filename = sigfile;
  t->unlink = 1; /* ok to remove this file after sending. */

  return a;
}


/* Encrypt the mail body A to all keys given as space separated keyids
   or fingerprints in KEYLIST and return the encrypted body.  */
static BODY *encrypt_message_pgp (BODY *a, char *keylist, int sign)
{
  char *outfile = NULL;
  BODY *t;
  GpgmeRecipients rset;
  GpgmeData plaintext;
  
  rset = create_recipient_set (keylist);
  if (!rset)
    return NULL;
  
  if (sign)
    convert_to_7bit (a);
  plaintext = body_to_data_object (a, 0);
  if (!plaintext)
    {
      gpgme_recipients_release (rset);
      return NULL;
    }
  
  outfile = encrypt_gpgme_object (plaintext, rset, 0, sign);
  gpgme_data_release (plaintext);
  gpgme_recipients_release (rset);
  if (!outfile)
      return NULL;

  t = mutt_new_body ();
  t->type = TYPEMULTIPART;
  t->subtype = safe_strdup ("encrypted");
  t->encoding = ENC7BIT;
  t->use_disp = 0;
  t->disposition = DISPINLINE;

  mutt_generate_boundary(&t->parameter);
  mutt_set_parameter("protocol", "application/pgp-encrypted", &t->parameter);
  
  t->parts = mutt_new_body ();
  t->parts->type = TYPEAPPLICATION;
  t->parts->subtype = safe_strdup ("pgp-encrypted");
  t->parts->encoding = ENC7BIT;

  t->parts->next = mutt_new_body ();
  t->parts->next->type = TYPEAPPLICATION;
  t->parts->next->subtype = safe_strdup ("octet-stream");
  t->parts->next->encoding = ENC7BIT;
  t->parts->next->filename = outfile;
  t->parts->next->use_disp = 1;
  t->parts->next->disposition = DISPINLINE;
  t->parts->next->unlink = 1; /* delete after sending the message */
  t->parts->next->d_filename = safe_strdup ("msg.asc"); /* non pgp/mime can save */

  return t;
}



/* Encrypt the mail body A to all keys given as space separated
   fingerprints in KEYLIST and return the S/MIME encrypted body.  */
static BODY *encrypt_message_smime (BODY *a, char *keylist)
{
  char *outfile = NULL;
  BODY *t;
  GpgmeRecipients rset;
  GpgmeData plaintext;

  rset = create_recipient_set (keylist);
  if (!rset)
    return NULL;

  plaintext = body_to_data_object (a, 0);
  if (!plaintext)
    {
      gpgme_recipients_release (rset);
      return NULL;
    }

  outfile = encrypt_gpgme_object (plaintext, rset, 1, 0);
  gpgme_data_release (plaintext);
  gpgme_recipients_release (rset);
  if (!outfile) 
      return NULL;

  t = mutt_new_body ();
  t->type = TYPEAPPLICATION;
  t->subtype = safe_strdup ("pkcs7-mime");
  mutt_set_parameter ("name", "smime.p7m", &t->parameter);
  mutt_set_parameter ("smime-type", "enveloped-data", &t->parameter);
  t->encoding = ENCBASE64;  /* The output of OpenSSL SHOULD be binary */
  t->use_disp = 1;
  t->disposition = DISPATTACH;
  t->d_filename = safe_strdup ("smime.p7m");
  t->filename = outfile;
  t->unlink = 1; /*delete after sending the message */
  t->parts=0;
  t->next=0;
  
  return t;
}



/* 
  Copy a clearsigned message, and strip the signature and PGP's
  dash-escaping.
  
  XXX - charset handling: We assume that it is safe to do
  character set decoding first, dash decoding second here, while
  we do it the other way around in the main handler.
  
  (Note that we aren't worse than Outlook & Cie in this, and also
  note that we can successfully handle anything produced by any
  existing versions of mutt.)  */

static void copy_clearsigned (GpgmeData data, STATE *s, char *charset)
{
  char buf[HUGE_STRING];
  short complete, armor_header;
  FGETCONV *fc;
  char *fname;
  FILE *fp;

  fname = data_object_to_tempfile (data, &fp);
  if (!fname)
    return;
  unlink (fname);
  FREE (fname);
  
  fc = fgetconv_open (fp, charset, Charset, M_ICONV_HOOK_FROM);
  
  for (complete = 1, armor_header = 1;
       fgetconvs (buf, sizeof (buf), fc) != NULL;
       complete = strchr (buf, '\n') != NULL)
  {
    if (!complete)
    {
      if (!armor_header)
	state_puts (buf, s);
      continue;
    }

    if (mutt_strcmp (buf, "-----BEGIN PGP SIGNATURE-----\n") == 0)
      break;
    
    if (armor_header)
    {
      if (buf[0] == '\n') 
	armor_header = 0;
      continue;
    }
    
    if (s->prefix) 
      state_puts (s->prefix, s);
    
    if (buf[0] == '-' && buf[1] == ' ')
      state_puts (buf + 2, s);
    else
      state_puts (buf, s);
  }
  
  fgetconv_close (&fc);
  fclose (fp);
}


#if 0
/* Return a string represenation of the signature status STATUS.  The
   return string ist a constant. */
static const char *
sig_status_string (GpgmeSigStat status)
{
  const char *s = "?";

  switch (status)
    {
    case GPGME_SIG_STAT_NONE:
      s = _("None");
      break;
    case GPGME_SIG_STAT_NOSIG:
      s = _("No Signature");
      break;
    case GPGME_SIG_STAT_GOOD:
    case GPGME_SIG_STAT_GOOD_EXP:
    case GPGME_SIG_STAT_GOOD_EXPKEY:
      s = _("Good");
      break;
    case GPGME_SIG_STAT_BAD:
      s = _("Bad");
      break;
    case GPGME_SIG_STAT_NOKEY:
      s = _("No Key");
      break;
    case GPGME_SIG_STAT_ERROR:
      s = _("Error");
      break;
    case GPGME_SIG_STAT_DIFF:
      s = _("More than one signature");
      break;
    }
  return s;
}
#endif


/* Display the common attributes of the signature summary SUM.
   Return 1 if there is is a severe warning.
 */
static int show_sig_summary (unsigned long sum,
                              GpgmeCtx ctx, GpgmeKey key, int idx,
                              STATE *s)
{
  int severe = 0;

  if ((sum & GPGME_SIGSUM_KEY_REVOKED))
    {
      state_attach_puts (_("Warning: One of the keys has been revoked\n"),s);
      severe = 1;
    }

  if ((sum & GPGME_SIGSUM_KEY_EXPIRED))
    { 
      time_t at = gpgme_key_get_ulong_attr (key, GPGME_ATTR_EXPIRE,
                                            NULL, 0);
      if (at)
        {
          state_attach_puts (_("Warning: The key used to create the "
                               "signature expired at: "), s);
          print_time (at , s);
          state_attach_puts ("\n", s);
        }
      else
        state_attach_puts (_("Warning: At least one certification key "
                             "has expired\n"), s);
    }

  if ((sum & GPGME_SIGSUM_SIG_EXPIRED))
    {
      state_attach_puts (_("Warning: The signature expired at: "),s);
      print_time ( gpgme_get_sig_ulong_attr
                   (ctx, idx, GPGME_ATTR_EXPIRE, 0), s);
      state_attach_puts ("\n", s);
    }

  if ((sum & GPGME_SIGSUM_KEY_MISSING))
    state_attach_puts (_("Can't verify due to a missing "
                         "key or certificate\n"), s);

  if ((sum & GPGME_SIGSUM_CRL_MISSING))
    {
      state_attach_puts (_("The CRL is not available\n"), s);
      severe = 1;
    }

  if ((sum & GPGME_SIGSUM_CRL_TOO_OLD))
    {
      state_attach_puts (_("Available CRL is too old\n"), s);
      severe = 1;
    }

  if ((sum & GPGME_SIGSUM_BAD_POLICY))
    state_attach_puts (_("A policy requirement was not met\n"), s);

  if ((sum & GPGME_SIGSUM_SYS_ERROR))
    {
      const char *t0, *t1;

      state_attach_puts (_("A system error occured"), s );

      /* Try to figure out some more detailed system error information. */
      t0 = gpgme_get_sig_string_attr (ctx, idx, GPGME_ATTR_ERRTOK, 0);
      t1 = gpgme_get_sig_string_attr (ctx, idx, GPGME_ATTR_ERRTOK, 1);
      if (t0 || t1)
        {
          state_attach_puts (": ", s);
          if (t0)
              state_attach_puts (t0, s);
          if (t1 && !(t0 && !strcmp (t0, t1)))
            {
              if (t0)
                state_attach_puts (",", s);
              state_attach_puts (t1, s);
            }
        }
      state_attach_puts ("\n", s);
    }

  return severe;
}


static void show_fingerprint (GpgmeKey key, STATE *state)
{
  const char *s;
  int i, is_pgp;
  char *buf, *p;
  const char *prefix = _("Fingerprint: ");

  if (!key)
    return;
  s = gpgme_key_get_string_attr (key, GPGME_ATTR_FPR, NULL, 0);
  if (!s)
    return;
  is_pgp = !gpgme_key_get_ulong_attr (key, GPGME_ATTR_TYPE, NULL, 0);

  buf = safe_malloc ( strlen (prefix) + strlen(s) * 4 + 2 );
  strcpy (buf, prefix);
  p = buf + strlen (buf);
  if (is_pgp && strlen (s) == 40)
    {  /* PGP v4 style formatted. */
      for (i=0; *s && s[1] && s[2] && s[3] && s[4]; s += 4, i++)
        {
          *p++ = s[0];
          *p++ = s[1];
          *p++ = s[2];
          *p++ = s[3];
          *p++ = ' ';
          if (i == 4)
            *p++ = ' ';
        }
    }
  else
    {
      for (i=0; *s && s[1] && s[2]; s += 2, i++)
        {
          *p++ = s[0];
          *p++ = s[1];
          *p++ = is_pgp? ' ':':';
          if (is_pgp && i == 7)
            *p++ = ' ';
        }
    }

  /* just in case print remaining odd digits */
  for (; *s; s++)
    *p++ = *s;
  *p++ = '\n';
  *p = 0;
  state_attach_puts (buf, state);
  free (buf);
}

/* Show the valididy of a key used for one signature. */
static void show_one_sig_validity (GpgmeCtx ctx, int idx, STATE *s)
{
  const char *txt;

  txt = _("INTERNAL ERROR: Unknown key validity\n");
  switch ( gpgme_get_sig_ulong_attr(ctx, idx, GPGME_ATTR_VALIDITY, 0))
    {
    case GPGME_VALIDITY_UNKNOWN:
      txt = _("WARNING: We have NO indication whether "
              "the key belongs to the person named "
              "as shown above\n");
      break;
    case GPGME_VALIDITY_UNDEFINED:
      break;
    case GPGME_VALIDITY_NEVER:
      txt = _("WARNING: The key does NOT BELONG to "
              "the person named as shown above\n");
      break;
    case GPGME_VALIDITY_MARGINAL:
      txt = _("WARNING: It is NOT certain that the key "
              "belongs to the person named as shown above\n");
      break;
    case GPGME_VALIDITY_FULL:
    case GPGME_VALIDITY_ULTIMATE:
      txt = NULL;
      break;
    }
  if (txt)
    state_attach_puts (txt, s);
}

/* Show information about one signature.  This fucntion is called with
   the context CTX of a sucessful verification operation and the
   enumerator IDX which should start at 0 and incremete for each
   call/signature. 

   Return values are: 0 for normal procession, 1 for a bad signature,
   2 for a signature with a warning or -1 for no more signature.  */
static int show_one_sig_status (GpgmeCtx ctx, int idx, STATE *s)
{
  GpgmeSigStat status;
  time_t created;
  const char *fpr, *uid;
  GpgmeKey key = NULL;
  int i, anybad = 0, anywarn = 0;
  unsigned int sum;

  fpr = gpgme_get_sig_status (ctx, idx, &status, &created);
  if (!fpr)
    return -1; /* no more signatures. */

  if (status != GPGME_SIG_STAT_GOOD)
    anybad = 1;

  if ( !gpgme_get_sig_key (ctx, idx, &key) )
    uid = gpgme_key_get_string_attr (key, GPGME_ATTR_USERID, NULL, 0);
  else
    uid = "[?]";

  /* Get the sig summary vector. */
  sum = gpgme_get_sig_ulong_attr (ctx, idx, GPGME_ATTR_SIG_SUMMARY, 0);

  if (!s || !s->fpout || !(s->flags & M_DISPLAY))
    ; /* No state information so no way to print anything. */
  else if ( (sum & GPGME_SIGSUM_GREEN) )
    {
      state_attach_puts (_("Good signature from: "), s);
      state_attach_puts (uid, s);
      state_attach_puts ("\n", s);
      for (i=1; (uid = gpgme_key_get_string_attr (key, GPGME_ATTR_USERID,
                                                  NULL, i)); i++)
        {
          if (gpgme_key_get_ulong_attr (key, GPGME_ATTR_UID_REVOKED, NULL, i))
              continue;
          state_attach_puts (_("                aka: "), s);
          state_attach_puts (uid, s);
          state_attach_puts ("\n", s);
        }
      state_attach_puts (_("            created: "), s);
      print_time (created, s);
      state_attach_puts ("\n", s);
      if (show_sig_summary (sum, ctx, key, idx, s))
        anywarn = 1;
      show_one_sig_validity (ctx, idx, s);
    }
  else if ((sum & GPGME_SIGSUM_RED))
    {
      state_attach_puts (_("*BAD* signature claimed to be from: "), s);
      state_attach_puts (uid, s);
      state_attach_puts ("\n", s);
      show_sig_summary (sum, ctx, key, idx, s);
    }
  else if (status == GPGME_SIG_STAT_GOOD && key
           && !gpgme_key_get_ulong_attr (key, GPGME_ATTR_TYPE, NULL, 0))
    { /* We can't decide (yellow) but this is a PGP key with a good
         signature, so we display what a PGP user expects: The name,
         fingerprint and the key validity (which is neither fully or
         ultimate). */
      state_attach_puts (_("Good signature from: "), s);
      state_attach_puts (uid, s);
      state_attach_puts ("\n", s);
      state_attach_puts (_("            created: "), s);
      print_time (created, s);
      state_attach_puts ("\n", s);
      show_one_sig_validity (ctx, idx, s);
      show_fingerprint (key,s);
      if (show_sig_summary (sum, ctx, key, idx, s))
        anywarn = 1;
    }
  else /* can't decide (yellow) */
    {
      state_attach_puts (_("Error checking signature"), s);
      state_attach_puts ("\n", s);
      show_sig_summary (sum, ctx, key, idx, s);
    }
  gpgme_key_release (key);

  return anybad? 1: anywarn? 2 : 0;
}


/* Do the actual verification step. With IS_SMIME set to true we
   assume S/MIME (surprise!) */
static int verify_one (BODY *sigbdy, STATE *s,
                       const char *tempfile, int is_smime)
{
  int badsig = -1;
  int anywarn = 0;
  int err;
  GpgmeCtx ctx;
  GpgmeData signature, message;
  GpgmeSigStat status;

  signature = file_to_data_object (s->fpin, sigbdy->offset, sigbdy->length);
  if (!signature)
    return -1;

  /* We need to tell gpgme about the encoding because the backend can't
     auto-detect plain base-64 encoding which is used by S/MIME. */
  if (is_smime)
    gpgme_data_set_encoding (signature, GPGME_DATA_ENCODING_BASE64);

  err = gpgme_data_new_from_file (&message, tempfile, 1);
  if (err) 
    {
      gpgme_data_release (signature);
      mutt_error ("error allocating data object: %s\n", gpgme_strerror (err));
      return -1;
    }
  ctx = create_gpgme_context (is_smime);

  /* Note: We don't need a current time output because GPGME avoids
     such an attack by separating the meta information from the
     data. */
  state_attach_puts (_("[-- Begin signature information --]\n"), s);

  err = gpgme_op_verify (ctx, signature, message, &status );
  mutt_need_hard_redraw ();
  if (err)
    {
      char buf[200];
      
      snprintf (buf, sizeof(buf)-1, 
                _("Error: verification failed: %s\n"),
                gpgme_strerror (err));
      state_attach_puts (buf, s);
    }
  else
    { /* Verification succeeded, see what the result is. */
      int res, idx;
      int anybad = 0;

      for(idx=0; (res = show_one_sig_status (ctx, idx, s)) != -1; idx++)
        {
          if (res == 1)
            anybad = 1;
          else if (res == 2)
            anywarn = 2;
        }
      if (!anybad)
        badsig = 0;
    }

  /* Show the notation data only if there was no bad signature.
     Notation data is currently only available collective for all
     signatures. */
  if (!badsig)
    {
      const char *nota;
      
      if ( (nota=gpgme_get_notation (ctx)) )
        {
          state_attach_puts ("*** Begin Notation ***\n", s);
          state_attach_puts (nota, s);
          if (*nota && nota[strlen (nota)-1] != '\n')
            state_attach_puts ("\n", s);
          state_attach_puts ("*** End Notation ***\n", s);
        }
    }

  gpgme_release (ctx);
  
  state_attach_puts (_("[-- End signature information --]\n\n"), s);
  dprint (1, (debugfile, "verify_one: returning %d.\n", badsig));
  
  return badsig? 1: anywarn? 2 : 0;
}


/* Decrypt a PGP or SMIME message (depending on the boolean flag
   IS_SMIME) with body A described further by state S.  Write
   plaintext out to file FPOUT and return a new body.  For PGP returns
   a flag in R_IS_SIGNED to indicate whether this is a combined
   encrypted and signed message, for S/MIME it returns true when it is
   not a encrypted but a signed message.  */
static BODY *decrypt_part (BODY *a, STATE *s, FILE *fpout, int is_smime,
                           int *r_is_signed)
{
  struct stat info;
  BODY *tattach;
  int err;
  GpgmeCtx ctx;
  GpgmeData ciphertext, plaintext;
  GpgmeSigStat sigstatus;
  int maybe_signed = 0;
  int anywarn = 0;

  if (r_is_signed)
    *r_is_signed = 0;

  ctx = create_gpgme_context (is_smime);

 restart:
  /* Make a data object from the body, create context etc. */
  ciphertext = file_to_data_object (s->fpin, a->offset, a->length);
  if (!ciphertext)
    return NULL;
  plaintext = create_gpgme_data ();
  sigstatus = GPGME_SIG_STAT_NONE;

  /* Do the decryption or the verification in case of the S/MIME hack. */
  if (!is_smime)
    err = gpgme_op_decrypt_verify (ctx, ciphertext, plaintext, &sigstatus);
  else if (maybe_signed)
    err = gpgme_op_verify (ctx, ciphertext, plaintext, &sigstatus);
  else
    err = gpgme_op_decrypt (ctx, ciphertext, plaintext);
  gpgme_data_release (ciphertext);
  if (err)
    {
      if (is_smime && !maybe_signed && err == GPGME_Decryption_Failed)
        { /* Check whether this might be a signed message despite what
             the mime header told us.  Retry then. */
          const char *tmp = gpgme_get_op_info (ctx, 0);
          if (tmp && strstr (tmp,
                             "<unsupportedAlgorithm>?</unsupportedAlgorithm"))
            {
              maybe_signed = 1;
              gpgme_data_release (plaintext);
              goto restart;
            }
        }
      mutt_need_hard_redraw ();
      if ((s->flags & M_DISPLAY))
        {
          char buf[200];
          
          snprintf (buf, sizeof(buf)-1, 
                    _("[-- Error: decryption failed: %s --]\n\n"),
                    gpgme_strerror (err));
          state_attach_puts (buf, s);
        }
      gpgme_data_release (plaintext);
      gpgme_release (ctx);
      return NULL;
  }
  mutt_need_hard_redraw ();

  /* Read the output from GPGME, and make sure to change CRLF to LF,
     otherwise read_mime_header has a hard time parsing the message.  */
  if (data_object_to_stream (plaintext, fpout))
    {
      gpgme_data_release (plaintext);
      gpgme_release (ctx);
      return NULL;
    }
  gpgme_data_release (plaintext);

  a->is_signed_data = 0;
  if (sigstatus != GPGME_SIG_STAT_NONE)
    {
      int res, idx;
      int anybad = 0;
      
      if (maybe_signed)
        a->is_signed_data = 1;
      if(r_is_signed)
        *r_is_signed = -1; /* A signature exists. */

      if ((s->flags & M_DISPLAY))
        state_attach_puts (_("[-- Begin signature "
                             "information --]\n"), s);
      for(idx=0; (res = show_one_sig_status (ctx, idx, s)) != -1; idx++)
        {
          if (res == 1)
            anybad = 1;
          else if (res == 2)
            anywarn = 1;
        }
      if (!anybad && idx && r_is_signed && *r_is_signed)
        *r_is_signed = anywarn? 2:1; /* Good signature. */
      
      if ((s->flags & M_DISPLAY))
        state_attach_puts (_("[-- End signature "
                             "information --]\n\n"), s);
    }
  gpgme_release (ctx); ctx = NULL;

  fflush (fpout);
  rewind (fpout);
  tattach = mutt_read_mime_header (fpout, 0);
  if (tattach)
    {
      /*
       * Need to set the length of this body part.
       */
      fstat (fileno (fpout), &info);
      tattach->length = info.st_size - tattach->offset;
      
      tattach->warnsig = anywarn;

      /* See if we need to recurse on this MIME part.  */
      mutt_parse_part (fpout, tattach);
    }

  return tattach;
}




/* Decrypt a PGP/MIME message in FPIN and B and return a new body and
   the stream in CUR and FPOUT.  Returns 0 on success. */
int pgp_decrypt_mime (FILE *fpin, FILE **fpout, BODY *b, BODY **cur)
{
  char tempfile[_POSIX_PATH_MAX];
  STATE s;
  BODY *first_part = b;
  int is_signed;
  
  first_part->goodsig = 0;
  first_part->warnsig = 0;

  if(!mutt_is_multipart_encrypted(b))
    return -1;

  if(!b->parts || !b->parts->next)
    return -1;
  
  b = b->parts->next;
  
  memset (&s, 0, sizeof (s));
  s.fpin = fpin;
  mutt_mktemp (tempfile);
  if (!(*fpout = safe_fopen (tempfile, "w+")))
  {
    mutt_perror (tempfile);
    return -1;
  }
  unlink (tempfile);

  *cur = decrypt_part (b, &s, *fpout, 0, &is_signed);
  rewind (*fpout);
  if (is_signed > 0)
    first_part->goodsig = 1;
  
  return *cur? 0:-1;
}


/* Decrypt a S/MIME message in FPIN and B and return a new body and
   the stream in CUR and FPOUT.  Returns 0 on success. */
int smime_decrypt_mime (FILE *fpin, FILE **fpout, BODY *b, BODY **cur)
{
  char tempfile[_POSIX_PATH_MAX];
  STATE s;
  FILE *tmpfp=NULL;
  int is_signed;
  long saved_b_offset;
  size_t saved_b_length;
  int saved_b_type;

  if (!mutt_is_application_smime (b))
    return -1;

  if (b->parts)
    return -1;
  
  /* Decode the body - we need to pass binary CMS to the
     backend.  The backend allows for Base64 encoded data but it does
     not allow for QP which I have seen in some messages.  So better
     do it here. */
  saved_b_type = b->type;
  saved_b_offset = b->offset;
  saved_b_length = b->length;
  memset (&s, 0, sizeof (s));
  s.fpin = fpin;
  fseek (s.fpin, b->offset, 0); 
  mutt_mktemp (tempfile);
  if (!(tmpfp = safe_fopen (tempfile, "w+")))
    {
      mutt_perror (tempfile);
      return -1;
    }
  mutt_unlink (tempfile);

  s.fpout = tmpfp;
  mutt_decode_attachment (b, &s);
  fflush (tmpfp);
  b->length = ftell (s.fpout);
  b->offset = 0;
  rewind (tmpfp);

  memset (&s, 0, sizeof (s));
  s.fpin = tmpfp;
  s.fpout = 0;
  mutt_mktemp (tempfile);
  if (!(*fpout = safe_fopen (tempfile, "w+")))
    {
      mutt_perror (tempfile);
      return -1;
    }
  mutt_unlink (tempfile);

  *cur = decrypt_part (b, &s, *fpout, 1, &is_signed);
  if (*cur)
    (*cur)->goodsig = is_signed > 0;
  b->type = saved_b_type;
  b->length = saved_b_length;
  b->offset = saved_b_offset;
  fclose (tmpfp);
  rewind (*fpout);
  if (*cur && !is_signed && !(*cur)->parts && mutt_is_application_smime (*cur))
    {
      /* Assume that this is a opaque signed s/mime message.  This is
         an ugly way of doing it but we have anyway a problem with
         arbitrary encoded S/MIME messages: Only the outer part may be
         encrypted.  The entire mime parsing should be revamped,
         probably by keeping the temportary files so that we don't
         need to decrypt them all the time.  Inner parts of an
         encrypted part can then pint into this file and tehre won't
         never be a need to decrypt again.  This needs a partial
         rewrite of the MIME engine. */
      BODY *bb = *cur;
      BODY *tmp_b;

      saved_b_type = bb->type;
      saved_b_offset = bb->offset;
      saved_b_length = bb->length;
      memset (&s, 0, sizeof (s));
      s.fpin = *fpout;
      fseek (s.fpin, bb->offset, 0); 
      mutt_mktemp (tempfile);
      if (!(tmpfp = safe_fopen (tempfile, "w+")))
        {
          mutt_perror (tempfile);
          return -1;
        }
      mutt_unlink (tempfile);

      s.fpout = tmpfp;
      mutt_decode_attachment (bb, &s);
      fflush (tmpfp);
      bb->length = ftell (s.fpout);
      bb->offset = 0;
      rewind (tmpfp);
      fclose (*fpout); 

      memset (&s, 0, sizeof (s));
      s.fpin = tmpfp;
      s.fpout = 0;
      mutt_mktemp (tempfile);
      if (!(*fpout = safe_fopen (tempfile, "w+")))
        {
          mutt_perror (tempfile);
          return -1;
        }
      mutt_unlink (tempfile);

      tmp_b = decrypt_part (bb, &s, *fpout, 1, &is_signed);
      if (tmp_b)
        tmp_b->goodsig = is_signed > 0;
      bb->type = saved_b_type;
      bb->length = saved_b_length;
      bb->offset = saved_b_offset;
      fclose (tmpfp);
      rewind (*fpout);
      mutt_free_body (cur);
      *cur = tmp_b;
    }
  return *cur? 0:-1;
}



/* Decrypt a message in FPIN and B and return a new body and the
   stream in CUR and FPOUT.  Returns 0 on success.  This fucntion
   decides for itself how the message is encrypted. */
int crypt_decrypt_mime (FILE *fpin, FILE **fpout, BODY *b, BODY **cur)
{
  int err;

  dprint (2, (debugfile, "Entering crypt_decrypt_mime\n"));
  if (mutt_is_application_smime (b))
    err = smime_decrypt_mime (fpin, fpout, b, cur);
  else if (mutt_is_multipart_encrypted (b))
    err = pgp_decrypt_mime (fpin, fpout, b, cur);
  else
    err = -1;

  dprint (2, (debugfile, "Leaving crypt_decrypt_mime\n"));
  return err;
}



/*
    Classic PGP support 
*/
static int pgp_check_traditional_one_body (FILE *fp, BODY *b, int tagged_only)
{
  char tempfile[_POSIX_PATH_MAX];
  char buf[HUGE_STRING];
  FILE *tfp;
  
  short sgn = 0;
  short enc = 0;
  
  if (b->type != TYPETEXT)
    return 0;

  if (tagged_only && !b->tagged)
    return 0;

  mutt_mktemp (tempfile);
  if (mutt_decode_save_attachment (fp, b, tempfile, 0, 0) != 0)
  {
    unlink (tempfile);
    return 0;
  }
  
  if ((tfp = fopen (tempfile, "r")) == NULL)
  {
    unlink (tempfile);
    return 0;
  }
  
  while (fgets (buf, sizeof (buf), tfp))
  {
    if (mutt_strncmp ("-----BEGIN PGP ", buf, 15) == 0)
    {
      if (mutt_strcmp ("MESSAGE-----\n", buf + 15) == 0)
	enc = 1;
      else if (mutt_strcmp ("SIGNED MESSAGE-----\n", buf + 15) == 0)
	sgn = 1;
    }
  }
  safe_fclose (&tfp);
  unlink (tempfile);

  if (!enc && !sgn)
    return 0;

  /* fix the content type */
  
  mutt_set_parameter ("format", "fixed", &b->parameter);
  mutt_set_parameter ("x-action", enc ? "pgp-encrypted" : "pgp-signed", &b->parameter);
  
  return 1;
}


int pgp_check_traditional (FILE *fp, BODY *b, int tagged_only)
{
  int rv = 0;
  int r;
  for (; b; b = b->next)
  {
    if (is_multipart (b))
      rv = pgp_check_traditional (fp, b->parts, tagged_only) || rv;
    else if (b->type == TYPETEXT)
    {
      if ((r = mutt_is_application_pgp (b)))
	rv = rv || r;
      else
	rv = pgp_check_traditional_one_body (fp, b, tagged_only) || rv;
    }
  }
  return rv;
}



/*
     Regular handler used by the main engine to format messages
*/



/* Support for multipart/signed.  */
void mutt_signed_handler (BODY *a, STATE *s)
{
  char tempfile[_POSIX_PATH_MAX]; 
  char *protocol;
  int protocol_major = TYPEOTHER;
  char *protocol_minor = NULL;
  
  BODY *b = a;
  BODY **signatures = NULL;
  int sigcnt = 0;
  int i;
  short goodsig = 1;

  dprint (2, (debugfile, "Entering mutt_signed_handler\n"));
  protocol = mutt_get_parameter ("protocol", a->parameter);
  a = a->parts;

  /* extract the protocol information */
  
  if (protocol)
    {
      char major[STRING];
      char *t;
      
      if ((protocol_minor = strchr (protocol, '/')))
        protocol_minor++;
      
      strfcpy (major, protocol, sizeof(major));
      if ((t = strchr(major, '/')))
        *t = '\0';
      
      protocol_major = mutt_check_mime_type (major);
    }

  /* consistency check */
  if (!(a && a->next && a->next->type == protocol_major && 
        !mutt_strcasecmp (a->next->subtype, protocol_minor)))
    {
      state_attach_puts (_("[-- Error: Inconsistent multipart/signed "
                           "structure! --]\n\n"), s);
      mutt_body_handler (a, s);
      return;
    }

  
  if (protocol_major == TYPEAPPLICATION
      && !mutt_strcasecmp (protocol_minor, "pgp-signature"))
    ;
  else if (protocol_major == TYPEAPPLICATION
           && !mutt_strcasecmp (protocol_minor, "pkcs7-signature"))
    ; 
  else if (protocol_major == TYPEAPPLICATION
           && !mutt_strcasecmp (protocol_minor, "x-pkcs7-signature"))
    ; 
  else if (protocol_major == TYPEMULTIPART
           && !mutt_strcasecmp (protocol_minor, "mixed"))
    ;
  else
    {
      state_printf (s, _("[-- Error: Unknown multipart/signed "
                         "protocol %s! --]\n\n"), protocol);
      mutt_body_handler (a, s);
      return;
    }
  
  if (s->flags & M_DISPLAY)
    {
      crypt_fetch_signatures (&signatures, a->next, &sigcnt);
      
      if (sigcnt)
        {
          mutt_mktemp (tempfile);
          if (crypt_write_signed (a, s, tempfile) == 0)
            {
              for (i = 0; i < sigcnt; i++)
                {
                  if (signatures[i]->type == TYPEAPPLICATION 
                      && !mutt_strcasecmp (signatures[i]->subtype,
                                           "pgp-signature"))
                    {
                      if (verify_one (signatures[i], s, tempfile, 0))
                        goodsig = 0;
                      continue;
                    }
                  if (signatures[i]->type == TYPEAPPLICATION 
                      && !mutt_strcasecmp (signatures[i]->subtype,
                                           "pkcs7-signature"))
                    {
                      if (verify_one (signatures[i], s, tempfile, 1))
                        goodsig = 0;
                      continue;
                    }
                  if (signatures[i]->type == TYPEAPPLICATION 
                      && !mutt_strcasecmp (signatures[i]->subtype,
                                           "x-pkcs7-signature"))
                    {
                      if (verify_one (signatures[i], s, tempfile, 1))
                        goodsig = 0;
                      continue;
                    }
                  state_printf (s, _("[-- Warning: We can't verify "
                                     "%s/%s signatures. --]\n\n"),
                                TYPE(signatures[i]), signatures[i]->subtype);
                }
            }
          
          mutt_unlink (tempfile);

          b->goodsig = goodsig;
      
          /* Now display the signed body */
          state_attach_puts (_("[-- The following data is signed --]\n\n"), s);
          FREE (signatures);
        }
      else
        state_attach_puts (_("[-- Warning: Can't find "
                             "any signatures. --]\n\n"), s);
    }
  
  mutt_body_handler (a, s);
  
  if (s->flags & M_DISPLAY && sigcnt)
    state_attach_puts (_("\n[-- End of signed data --]\n"), s);
  dprint (2, (debugfile, "Leaving mutt_signed_handler\n"));
}


/* Support for classic_application/pgp */
void pgp_application_pgp_handler (BODY *m, STATE *s)
{
  int needpass = -1, pgp_keyblock = 0;
  int clearsign = 0;
  long start_pos = 0;
  long bytes, last_pos, offset;
  char buf[HUGE_STRING];
  FILE *pgpout = NULL;

  GpgmeError err;
  GpgmeData armored_data = NULL;

  short maybe_goodsig = 1;
  short have_any_sigs = 0;

  char body_charset[STRING];

  dprint (2, (debugfile, "Entering pgp_application_pgp handler\n"));
  mutt_get_body_charset (body_charset, sizeof (body_charset), m);
  
  fseek (s->fpin, m->offset, 0);
  last_pos = m->offset;
  
  for (bytes = m->length; bytes > 0;)
    {
      if (fgets (buf, sizeof (buf), s->fpin) == NULL)
        break;
      
      offset = ftell (s->fpin);
      bytes -= (offset - last_pos); /* don't rely on mutt_strlen(buf) */
      last_pos = offset;
      
      if (mutt_strncmp ("-----BEGIN PGP ", buf, 15) == 0)
        {
          clearsign = 0;
          start_pos = last_pos;
          
          if (mutt_strcmp ("MESSAGE-----\n", buf + 15) == 0)
            needpass = 1;
          else if (mutt_strcmp ("SIGNED MESSAGE-----\n", buf + 15) == 0)
            {
              clearsign = 1;
              needpass = 0;
            }
          else if (!option (OPTDONTHANDLEPGPKEYS) &&
                   mutt_strcmp ("PUBLIC KEY BLOCK-----\n", buf + 15) == 0)
            {
              needpass = 0;
              pgp_keyblock =1;
            } 
          else
            {
              /* XXX - we may wish to recode here */
              if (s->prefix)
                state_puts (s->prefix, s);
              state_puts (buf, s);
              continue;
            }
          
          have_any_sigs = have_any_sigs || (clearsign && (s->flags & M_VERIFY));
          
          /* Copy PGP material to an data container */
          armored_data = create_gpgme_data ();
          gpgme_data_write (armored_data, buf, strlen (buf));
          while (bytes > 0 && fgets (buf, sizeof (buf) - 1, s->fpin) != NULL)
            {
              offset = ftell (s->fpin);
              bytes -= (offset - last_pos); /* don't rely on mutt_strlen(buf) */
              last_pos = offset;
              
              gpgme_data_write (armored_data, buf, strlen (buf));
              
              if ((needpass && !mutt_strcmp ("-----END PGP MESSAGE-----\n", buf)) ||
                  (!needpass 
                   && (!mutt_strcmp ("-----END PGP SIGNATURE-----\n", buf)
                       || !mutt_strcmp ("-----END PGP PUBLIC KEY BLOCK-----\n",buf))))
                break;
            }
          
          /* Invoke PGP if needed */
          if (!clearsign || (s->flags & M_VERIFY))
            {
              GpgmeData plaintext;
              GpgmeCtx ctx;
              GpgmeSigStat sigstatus;
              
              plaintext = create_gpgme_data ();
              ctx = create_gpgme_context (0);

              if (clearsign)
                err = gpgme_op_verify (ctx, armored_data, plaintext,
                                       &sigstatus);
              else
                {
                  err = gpgme_op_decrypt_verify (ctx, armored_data, plaintext,
                                                 &sigstatus);
                  if (err == GPGME_No_Data)
                    {
                      /* Decrypt verify can't handle signed only messages. */
                      gpgme_data_rewind (armored_data);
                      /* Must release plaintext so that we supply an
                         uninitialized object. */
                      gpgme_data_release (plaintext);
                      plaintext = create_gpgme_data ();
                      err = gpgme_op_verify (ctx, armored_data, plaintext,
                                             &sigstatus);
                    }
                }
              if (err)
                {
                  char errbuf[200];
                  
                  snprintf (errbuf, sizeof(errbuf)-1, 
                            _("Error: decryption/verification failed: %s\n"),
                            gpgme_strerror (err));
                  state_attach_puts (errbuf, s);
                }
              else
                { /* Decryption/Verification succeeded */
                  char *tmpfname;

                  have_any_sigs = 0;
                  maybe_goodsig = 0;
                  if ( (s->flags & M_DISPLAY)
                       && sigstatus != GPGME_SIG_STAT_NONE)
                    {
                      int res, idx;
                      int anybad = 0;
                      int anywarn = 0;

                      state_attach_puts (_("[-- Begin signature "
                                           "information --]\n"), s);
                      have_any_sigs = 1;
                      for(idx=0;
                          (res = show_one_sig_status (ctx, idx, s)) != -1;
                          idx++)
                        {
                          if (res == 1)
                            anybad = 1;
                          else if (res == 2)
                            anywarn = 1;
                        }
                      if (!anybad && idx)
                        maybe_goodsig = 1;
                      
                      state_attach_puts (_("[-- End signature "
                                           "information --]\n\n"), s);
                    }

                  tmpfname = data_object_to_tempfile (plaintext, &pgpout);
                  if (!tmpfname)
                    {
                      pgpout = NULL;
                      state_attach_puts (_("Error: copy data failed\n"), s);
                    }
                  else
                    {
                      unlink (tmpfname);
                      FREE (tmpfname);
                    }
                }
              gpgme_release (ctx);
            }
      
          /*
           * Now, copy cleartext to the screen.  NOTE - we expect that PGP
           * outputs utf-8 cleartext.  This may not always be true, but it 
           * seems to be a reasonable guess.
           */
          
          if(s->flags & M_DISPLAY)
            {
              if (needpass)
            state_attach_puts (_("[-- BEGIN PGP MESSAGE --]\n\n"), s);
              else if (pgp_keyblock)
                state_attach_puts (_("[-- BEGIN PGP PUBLIC KEY BLOCK --]\n"), s);
              else
                state_attach_puts (_("[-- BEGIN PGP SIGNED MESSAGE --]\n\n"), s);
            }
          
          if (clearsign)
            {
              copy_clearsigned (armored_data, s, body_charset);
            }
          else if (pgpout)
            {
              FGETCONV *fc;
              int c;
              rewind (pgpout);
              fc = fgetconv_open (pgpout, "utf-8", Charset, 0);
              while ((c = fgetconv (fc)) != EOF)
                {
                  state_putc (c, s);
                  if (c == '\n' && s->prefix)
                    state_puts (s->prefix, s);
                }
              fgetconv_close (&fc);
            }
          
          if (s->flags & M_DISPLAY)
            {
              state_putc ('\n', s);
              if (needpass)
                state_attach_puts (_("[-- END PGP MESSAGE --]\n"), s);
              else if (pgp_keyblock)
                state_attach_puts (_("[-- END PGP PUBLIC KEY BLOCK --]\n"), s);
              else
                state_attach_puts (_("[-- END PGP SIGNED MESSAGE --]\n"), s);
            }
          
          if (pgpout)
            {
              safe_fclose (&pgpout);
            }
        }
      else
        {
          /* XXX - we may wish to recode here */
          if (s->prefix)
            state_puts (s->prefix, s);
          state_puts (buf, s);
        }
    }

  m->goodsig = (maybe_goodsig && have_any_sigs);
  
  if (needpass == -1)
    {
      state_attach_puts (_("[-- Error: could not find beginning of PGP message! --]\n\n"), s);
      return;
    }
  dprint (2, (debugfile, "Leaving pgp_application_pgp handler\n"));
}


/* MIME handler for pgp/mime encrypted messages. */
void pgp_encrypted_handler (BODY *a, STATE *s)
{
  char tempfile[_POSIX_PATH_MAX];
  FILE *fpout;
  BODY *tattach;
  BODY *orig_body = a;
  int is_signed;
  
  dprint (2, (debugfile, "Entering pgp_encrypted handler\n"));
  a = a->parts;
  if (!a || a->type != TYPEAPPLICATION || !a->subtype
      || ascii_strcasecmp ("pgp-encrypted", a->subtype) 
      || !a->next || a->next->type != TYPEAPPLICATION || !a->next->subtype
      || ascii_strcasecmp ("octet-stream", a->next->subtype) )
    {
      if (s->flags & M_DISPLAY)
        state_attach_puts (_("[-- Error: malformed PGP/MIME message! --]\n\n"),
                           s);
      return;
    }

  /* Move forward to the application/pgp-encrypted body. */
  a = a->next;

  mutt_mktemp (tempfile);
  if (!(fpout = safe_fopen (tempfile, "w+")))
    {
      if (s->flags & M_DISPLAY)
        state_attach_puts (_("[-- Error: could not create temporary file! "
                             "--]\n"), s);
      return;
    }

  tattach = decrypt_part (a, s, fpout, 0, &is_signed);
  if (tattach)
    {
      tattach->goodsig = is_signed > 0;

      if (s->flags & M_DISPLAY)
        state_attach_puts (is_signed?
          _("[-- The following data is PGP/MIME signed and encrypted --]\n\n"):
          _("[-- The following data is PGP/MIME encrypted --]\n\n"),
                           s);
      
      {
        FILE *savefp = s->fpin;
        s->fpin = fpout;
        mutt_body_handler (tattach, s);
        s->fpin = savefp;
      }

      /* 
       * if a multipart/signed is the _only_ sub-part of a
       * multipart/encrypted, cache signature verification
       * status.
       */
      if (mutt_is_multipart_signed (tattach) && !tattach->next)
        orig_body->goodsig |= tattach->goodsig;
    
      if (s->flags & M_DISPLAY)
        {
          state_puts ("\n", s);
          state_attach_puts (is_signed?
             _("[-- End of PGP/MIME signed and encrypted data --]\n"):
             _("[-- End of PGP/MIME encrypted data --]\n"),
                             s);
        }

      mutt_free_body (&tattach);
    }
  
  fclose (fpout);
  mutt_unlink(tempfile);
  dprint (2, (debugfile, "Leaving pgp_encrypted handler\n"));
}

/* Support for application/smime */
void smime_encrypted_handler (BODY *a, STATE *s)
{
  char tempfile[_POSIX_PATH_MAX];
  FILE *fpout;
  BODY *tattach;
  int is_signed;


  dprint (2, (debugfile, "Entering smime_encrypted handler\n"));
  
  a->warnsig = 0;
  mutt_mktemp (tempfile);
  if (!(fpout = safe_fopen (tempfile, "w+")))
    {
      if (s->flags & M_DISPLAY)
        state_attach_puts (_("[-- Error: could not create temporary file! "
                             "--]\n"), s);
      return;
    }

  tattach = decrypt_part (a, s, fpout, 1, &is_signed);
  if (tattach)
    {
      tattach->goodsig = is_signed > 0;

      if (s->flags & M_DISPLAY)
        state_attach_puts (is_signed?
          _("[-- The following data is S/MIME signed --]\n\n"):
          _("[-- The following data is S/MIME encrypted --]\n\n"),
                           s);
      
      {
        FILE *savefp = s->fpin;
        s->fpin = fpout;
        mutt_body_handler (tattach, s);
        s->fpin = savefp;
      }

      /* 
       * if a multipart/signed is the _only_ sub-part of a
       * multipart/encrypted, cache signature verification
       * status.
       */
      if (mutt_is_multipart_signed (tattach) && !tattach->next)
        {
          if (!(a->goodsig = tattach->goodsig))
            a->warnsig = tattach->warnsig;
        }
      else if (tattach->goodsig)
        {
          a->goodsig = 1;
          a->warnsig = tattach->warnsig;
        }

      if (s->flags & M_DISPLAY)
        {
          state_puts ("\n", s);
          state_attach_puts (is_signed?
             _("[-- End of S/MIME signed data --]\n"):
             _("[-- End of S/MIME encrypted data --]\n"),
                             s);
        }

      mutt_free_body (&tattach);
    }
  
  fclose (fpout);
  mutt_unlink(tempfile);
  dprint (2, (debugfile, "Leaving smime_encrypted handler\n"));
}



#endif /*HAVE_GPGME*/
